@@ -7,18 +7,23 @@ import platform
 import signal
 import logging
 from PyQt5 import QtCore
-from threading import Event, Thread, current_thread
+from threading import Event, Thread, current_thread, activeCount
 import webview.constant as constant
 from cefpython3 import cefpython as cef
 from uuid import uuid4
 from PyQt5.QtGui import *
 from PyQt5.QtCore import *
 from PyQt5.QtWidgets import *
+import hashlib
+import webbrowser
 
+splash = None  # 闪屏界面
 third_party_client_pid_lists = []  # 记录应用生命周期中所启动的第三方应用程序的PID
+third_party_client_launch_status = {}  # 记录每个第三方应用当前启动状态{ -1: 未启动,  0: 启动中,  1: 已启动 }
 pixel_ratio = 1
 screen_width = 0
 screen_height = 0
+qt_class_name = constant.qt_class_name
 default_window_width = constant.default_window_width
 default_window_height = constant.default_window_height
 default_window_title = constant.default_window_title
@@ -40,12 +45,23 @@ dpi_dict = {
     '96': 1,
     '120': 1.25,
     '144': 1.5,
+    '168': 1.75,
     '192': 2
 }
 
 
+# 打点记录每个页面加载情况
+def record_page_monitor(browser, data):
+    global app_inject_obj
+    global app_cef_cache_path
+    if 'BuriedPoint' in app_inject_obj.keys():
+        app_inject_obj['BuriedPoint'].add_embed_page_operate(data[0])
+    if app_cef_cache_path != '':
+        browser.ExecuteFunction('window.python_cef.increaseMonitor', data)
+
+
 class Dialog(QDialog):
-    def __init__(self, params={}):
+    def __init__(self, params={}, view=None):
         super(Dialog, self).__init__()
         default_params = {
             'topBgColor': '#2a5596',
@@ -60,19 +76,22 @@ class Dialog(QDialog):
             'dialogHeight': 201,
             'leftButtonText': 'Left Button',
             'rightButtonText': 'Right Button',
-            'leftButtonAction': 'close',
+            'leftButtonAction': 'emit',
             'rightButtonAction': 'cancel',
             'buttonWidth': 110,
             'buttonHeight': 34,
             'buttonFontSize': 16,
             'borderRadius': 0,
-            'blurRadius': 20
+            'blurRadius': 20,
+            'customizedData': {}
         }
         default_params.update(params)
         global pixel_ratio
         if platform.system() == 'Windows':
             pixel_ratio = dpi_dict[str(self.logicalDpiX())]
             default_params['borderRadius'] = 6
+        self.setAttribute(Qt.WA_DeleteOnClose)
+        self.view = view
         self.pixel_ratio = pixel_ratio
         self.m_drag = False
         self.m_DragPosition = 0
@@ -87,7 +106,8 @@ class Dialog(QDialog):
     def init(self):
         action = {
             'close': self.quit_app,
-            'cancel': self.close
+            'cancel': self.close,
+            'emit': self.emit_event_to_browser
         }
         width = self.params['dialogWidth'] * self.pixel_ratio + self.params['blurRadius'] * 2 * self.pixel_ratio
         height = self.params['dialogHeight'] * self.pixel_ratio + self.params['blurRadius'] * 2 * self.pixel_ratio
@@ -222,6 +242,12 @@ class Dialog(QDialog):
         style = style.replace('[buttonHoverBgColor]', self.params['buttonHoverBgColor'])
         self.setStyleSheet(style)
 
+    def emit_event_to_browser(self):
+        if self.view:
+            self.view.dispatch_customize_event(event_name='cefDialogConfirmEvent',
+                                               event_data=self.params['customizedData'])
+        self.close()
+
     def mousePressEvent(self, event):
         if event.button() == Qt.LeftButton:
             self.m_drag = True
@@ -264,40 +290,65 @@ class CefApplication(QApplication):
         self.timer.stop()
 
 
+class PopUp(object):
+    def __init__(self, browser):
+        self.browser = browser
+
+    def closePopUp(self):
+        self.browser.CloseBrowser()
+
+
 class LoadHandler(object):
     def __init__(self, uid, payload, cid, browser):
         self.payload = payload
         self.uid = uid
         self.cid = cid
         self.browser = browser
-        self.main_frame_load_start_time = 0
-        self.main_frame_load_end_time = 0
+        self.handlerId = generate_guid()
 
     def OnLoadStart(self, browser, frame):
+        logging.debug(
+            'OnLoadStart: url = {url}'.format(url=browser.GetUrl()))
+        if browser.IsPopup():
+            bindings = cef.JavascriptBindings(bindToFrames=False, bindToPopups=False)
+            bindings.SetProperty('qt5CefVersion', constant.qt5_cef_version)
+            bindings.SetObject('cef', PopUp(browser))
+            browser.SetJavascriptBindings(bindings)
+
         if frame.IsMain():
             if debug_mode:
                 with open(os.path.dirname(__file__) + '/burgeon.cef.sdk.js', 'r', encoding='UTF-8') as js:
                     browser.ExecuteJavascript(js.read())
             else:
+                logging.debug('Inject cef sdk to cid={cid}, uid={uid}'.format(cid=self.cid, uid=self.uid))
                 browser.ExecuteJavascript(cef_sdk)
-            self.main_frame_load_start_time = time.time()
-            # set_javascript_bindings(self.uid)
-            append_payload(self.uid, self.payload, self.cid)
-            browser.ExecuteJavascript('window.__cef__.CEF_INFO.start_load_timestamp = {start_load_timestamp}'.format(
-                start_load_timestamp=self.main_frame_load_start_time))
-            browser.ExecuteJavascript(
-                'window.__cef__.CEF_INFO.language = "{language}"'.format(language=get_system_language()))
-            self.browser.update_browser_info_one_by_one()
+                append_payload(self.uid, self.payload, self.cid)
+                self.browser.update_browser_info_one_by_one()
+
+        record_page_monitor(browser, [{
+            'ID': hashlib.md5(
+                '{handlerId}{globalId}{Id}'.format(handlerId=self.handlerId, globalId=frame.GetBrowserIdentifier(),
+                                                   Id=frame.GetIdentifier()).encode(encoding='UTF-8')).hexdigest(),
+            'uid': self.uid,
+            'cid': self.cid,
+            'time': time.time() * 1000,
+            'url': frame.GetUrl(),
+            'loadStart': True,
+            'loadEnd': False
+        }])
 
     def OnLoadEnd(self, browser, frame, http_code):
-        if frame.IsMain():
-            # set_javascript_bindings(self.uid)
-            self.main_frame_load_end_time = time.time()
-            load_time_cost = self.main_frame_load_end_time - self.main_frame_load_start_time
-            browser.ExecuteJavascript('window.__cef__.CEF_INFO.end_load_timestamp = {start_load_timestamp}'.format(
-                start_load_timestamp=self.main_frame_load_end_time))
-            browser.ExecuteJavascript('window.__cef__.CEF_INFO.loadTimeCost = {time_cost}'.format(
-                time_cost=load_time_cost))
+        record_page_monitor(browser, [{
+            'ID': hashlib.md5(
+                '{handlerId}{globalId}{Id}'.format(handlerId=self.handlerId, globalId=frame.GetBrowserIdentifier(),
+                                                   Id=frame.GetIdentifier()).encode(encoding='UTF-8')).hexdigest(),
+            'uid': self.uid,
+            'cid': self.cid,
+            'time': time.time() * 1000,
+            'url': frame.GetUrl(),
+            'loadStart': False,
+            'loadEnd': True
+        }])
 
     def OnLoadError(self, browser, frame, error_code, error_text_out, failed_url):
         if frame.IsMain():
@@ -306,9 +357,23 @@ class LoadHandler(object):
                     browser.ExecuteJavascript(js.read())
             else:
                 browser.ExecuteJavascript(cef_sdk)
-            # set_javascript_bindings(self.uid)
-            append_payload(self.uid, self.payload, self.cid)
-            self.browser.update_browser_info_one_by_one()
+                append_payload(self.uid, self.payload, self.cid)
+                self.browser.update_browser_info_one_by_one()
+
+
+class LifeSpanHandler(object):
+    def __init__(self, validate_function=lambda *args: False):
+        self.validate_function = validate_function
+
+    def OnBeforePopup(self, browser, frame, target_url, target_frame_name, target_disposition, user_gesture,
+                      popup_features, window_info_out, client, browser_settings_out, no_javascript_access_out):
+        try:
+            if hasattr(self.validate_function, '__call__') and self.validate_function(target_url):
+                webbrowser.open(target_url)
+                return True
+            pass
+        except Exception as e:
+            logging.debug(e)
 
 
 class BrowserView(QMainWindow):
@@ -327,13 +392,14 @@ class BrowserView(QMainWindow):
                  height=default_window_height,
                  resizable=True, full_screen=False, min_size=(min_window_width, min_window_height),
                  background_color="#ffffff", web_view_ready=None, cid='', enable_max=True, window_type='cef',
-                 quit_app_when_main_window_close=True):
+                 quit_app_when_main_window_close=True, popup_validate=None):
         super(BrowserView, self).__init__()
         BrowserView.instances[uid] = self
         screen = QDesktopWidget().screenGeometry()
         global screen_width, screen_height
         screen_width = screen.width()
         screen_height = screen.height()
+        self.popup_validate = popup_validate
         self.quit_app_when_main_window_close = quit_app_when_main_window_close
         self.key_board_layout = 0  # 用于记录当前窗口在禁用输入法之前的状态
         self.attached_child_list = []  # 存储该窗口的跟随子窗口列表
@@ -419,6 +485,10 @@ class BrowserView(QMainWindow):
     def exit_application(self):
         quit_application()
 
+    def close_attached_window(self):
+        for child_window in self.attached_child_list:
+            child_window.close()
+
     def showEvent(self, event):
         if self.cid in BrowserView.cid_hwnd_map:
             parent = self.parent()
@@ -444,6 +514,7 @@ class BrowserView(QMainWindow):
                 self.view.ExecuteFunction('window.python_cef.dispatchCustomEvent', 'windowCloseEvent')
             else:
                 self.kill_subprocess()
+                self.close_attached_window()
                 event.accept()
         else:
             if self.uid == 'master':
@@ -452,13 +523,17 @@ class BrowserView(QMainWindow):
                     quit_application()
                 else:
                     self.kill_subprocess()
+                    self.close_attached_window()
                     self.update_browser_info_one_by_one(increase=False)
                     event.accept()
 
             else:
                 self.kill_subprocess()
+                self.close_attached_window()
                 self.update_browser_info_one_by_one(increase=False)
                 event.accept()
+        # print('[ closeEvent ]', self.cid, self.uid)
+        # print('[ closeEvent ] threading.currentThread() = ', current_thread().ident)
 
     def resizeEvent(self, event):
         cef.WindowUtils.OnSize(self.winId(), 0, 0, 0)
@@ -692,6 +767,7 @@ class BrowserView(QMainWindow):
         launch_third_party_application(param)
 
     def nest_frame_window(self, param={}):
+        logging.debug('[Nest Frame Window] : param = {param}'.format(param=param))
         if param is None:
             param = {}
         if isinstance(param, dict):
@@ -705,7 +781,8 @@ class BrowserView(QMainWindow):
             param.setdefault('left', default_nest_window_margin)  # 内嵌窗口距离target窗口的左侧距离
             param.setdefault('windowType', 'cef')
             frame_window = create_qt_view(url=param['url'], cid=param['newCid'], default_show=False,
-                                          payload=param['payload'], window_type=param['windowType'])
+                                          payload=param['payload'], window_type=param['windowType'],
+                                          popup_validate=self.popup_validate)
             frame_window.responsive_params['top'] = param['top']
             frame_window.responsive_params['right'] = param['right']
             frame_window.responsive_params['bottom'] = param['bottom']
@@ -737,11 +814,11 @@ class BrowserView(QMainWindow):
             'window.__cef__.CEF_INFO.windowLogicalHeight = {windowLogicalHeight}'.format(
                 windowLogicalHeight=self.height()))
 
-    def show_close_dialog(self, params={}):
+    def show_cef_dialog(self, params={}):
         if params is None:
             params = {}
         params['parentWindow'] = self
-        Dialog(params)
+        Dialog(params, self)
 
     def send_key_event_to_client(self, params={}):
         if params is None:
@@ -786,7 +863,6 @@ class BrowserView(QMainWindow):
         if platform.system() == 'Windows':
             import win32api
             import win32con
-            print('current ime : ', win32api.GetKeyboardLayout(), win32api.GetKeyboardLayoutName())
 
     def set_menu_info(self, menu='No Value'):
         BrowserView.menu_map[self.uid] = menu
@@ -809,8 +885,13 @@ class BrowserView(QMainWindow):
     def reload_ignore_cache(self):
         self.view.ReloadIgnoreCache()
 
-    def show_dev_tool(self):
-        self.view.ShowDevTools()
+    def show_dev_tool(self, cid=None):
+        if cid is None:
+            self.view.ShowDevTools()
+        else:
+            uid = self.get_uid_by_cid(cid)
+            if uid in BrowserView.instances.keys():
+                BrowserView.instances[uid].view.ShowDevTools()
 
     def set_window_title(self, title=default_window_title):
         self.setWindowTitle(title)
@@ -818,10 +899,28 @@ class BrowserView(QMainWindow):
     def reload_url_by_cid(self, cid, url):
         uid = self.get_uid_by_cid(cid)
         if uid in BrowserView.instances.keys():
-            BrowserView.instances[uid].view.ExecuteFunction('window.python_cef.loadUrl', url)
+            BrowserView.instances[uid].view.LoadUrl(url)
 
     def close_process(self, pid):
-        os.popen('taskkill /pid {pid} -f'.format(pid=pid))
+        # print(time.time(), ':',  '【try to close process pid = ', pid, '】')
+        try:
+            # os.popen('taskkill /pid {pid} -f'.format(pid=pid))
+            os.kill(pid, 0)
+            os.kill(pid, signal.SIGTERM)
+        except OSError as e:
+            # print(pid, ' is not exist')
+            logging.debug('{pid} is not exist'.format(pid=pid))
+
+    def load_url(self, url='about:blank'):
+        if hasattr(self, 'view'):
+            self.view.LoadUrl(url)
+
+    def reload_sdk(self):
+        logging.debug('[Reload SDK]')
+        self.view.ExecuteJavascript('console.log("%c U will reload cef sdk.", "font-weight: 600;color: #31136a")')
+        self.view.ExecuteJavascript('window.CEF_HAS_INITIALIZED = false;')
+        self.view.ExecuteJavascript(cef_sdk)
+        self.update_browser_info_one_by_one()
 
 
 class ThirdPartyWindow(QWidget):
@@ -841,7 +940,8 @@ def update_third_party_client_info(cid='un_know_client_cid', info={}):
                                                             cid, info)
 
 
-def listen_process_by_hwnd(hwnd, cid, client_window, pid):
+def listen_process_by_hwnd(hwnd, cid, client_window, pid, application_path):
+    # print(time.time(), ':',  'listen pid = ', pid)
     if hwnd == 0:
         return
     else:
@@ -849,16 +949,28 @@ def listen_process_by_hwnd(hwnd, cid, client_window, pid):
             import win32event
             try:
                 if win32event.WaitForSingleObject(hwnd, win32event.INFINITE) == win32event.WAIT_OBJECT_0:
-                    client_window.close()
+                    app_tag = hashlib.md5(application_path.encode(encoding='UTF-8')).hexdigest()
+                    third_party_client_launch_status[app_tag] = -1
                     update_third_party_client_info(cid, {'pid': pid, 'status': 'closed'})  # 在所有浏览器窗口中更新某个第三方应用的打开状态
+                    # print(pid, ' is be killed')
+                    # client_window.close()  # 不再执行窗口关闭动作
+                    BrowserView.instances['master'].view.ExecuteFunction('window.__cef__.close', [cid])
+                    # print(client_window.cid, 'is closed')
+                    # print('--------------- end ------------------', pid)
             except Exception as e:
                 logging.error(e)
 
 
-def find_main_window_hwnd(class_name=None, title=""):
+def find_main_window_hwnd(class_name=qt_class_name, title=""):
     if platform.system() == 'Windows':
         import win32gui
-        return win32gui.FindWindow(class_name, title)
+        hwnd = win32gui.FindWindow(class_name, title)
+        logging.debug(
+            '[ pid = {pid} ]  FindWindow class_name = {class_name}, title = {title}'.format(class_name=class_name,
+                                                                                            title=title,
+                                                                                            pid=os.getpid()))
+        logging.debug('[ pid = {pid} ] FindWindow hwnd = {hwnd}'.format(hwnd=hwnd, pid=os.getpid()))
+        return hwnd
     else:
         return 0
 
@@ -892,6 +1004,7 @@ def validate_3rd_client_alive(third_party_application_title):
 
 
 def get_handle_id(third_party_application_title):
+    logging.debug('[Get Client Handle Id]')
     if platform.system() == 'Windows':
         import win32gui
         hwnd = 0
@@ -927,6 +1040,7 @@ def launch_third_party_client(application_title,
                               target_window,
                               client_window,
                               cid):
+    logging.debug('[Launch Third Party Client] target_window cid = {cid}'.format(cid=target_window.cid))
     if platform.system() == 'Windows':
         import win32api
         import win32con
@@ -934,9 +1048,11 @@ def launch_third_party_client(application_title,
         for (k, v) in launch_params.items():
             exe_path = exe_path + ' ' + '--' + k + ':' + v + ''
         child_process = subprocess.Popen(exe_path)
+        logging.debug('[Launch Third Party Client] Client Pid = {pid}'.format(pid=child_process.pid))
         child_process_hwnd = win32api.OpenProcess(win32con.PROCESS_ALL_ACCESS, False, child_process.pid)
         update_third_party_client_info(cid, {'pid': child_process.pid, 'status': 'alive'})  # 处理第三方应用在所有窗口的进程状态
-        t = Thread(target=listen_process_by_hwnd, args=(child_process_hwnd, cid, client_window, child_process.pid))
+        t = Thread(target=listen_process_by_hwnd,
+                   args=(child_process_hwnd, cid, client_window, child_process.pid, application_path))
         t.start()
         BrowserView.cid_3rd_client_process_map[cid] = child_process.pid
         third_party_client_pid_lists.append(child_process.pid)
@@ -949,10 +1065,16 @@ def nest_third_party_application(target_uid='master',
                                  application_path='',
                                  launch_params={},
                                  show=True):
-    # third_party_application_title = generate_guid('third_party_application_title')
+    logging.debug('[Nest Third Party Application]')
+    app_tag = hashlib.md5(application_path.encode(encoding='UTF-8')).hexdigest()
+    if not third_party_client_launch_status.__contains__(app_tag):
+        third_party_client_launch_status[app_tag] = 0  # 启动中
+    elif third_party_client_launch_status[app_tag] != -1:
+        # 如果某个第三方应用已经启动或正在启动中，则不执行启动程序
+        return
+
     third_party_application_title = 'third_party_application_title_' + os.path.basename(application_path)  # 窗口名使用应用程序名
     third_party_wrapper_window = create_qt_view(default_show=False, window_type='qt', cid=cid)
-    third_party_wrapper_window.update_browser_info_one_by_one()
     global pixel_ratio
     if platform.system() == 'Windows':
         pixel_ratio = dpi_dict[str(third_party_wrapper_window.logicalDpiX())]
@@ -964,11 +1086,13 @@ def nest_third_party_application(target_uid='master',
     offset_left = pixel_ratio * geometry['left']
     t = Thread(target=launch_third_party_client,
                args=(third_party_application_title, application_path, launch_params, target_window,
-                     third_party_wrapper_window, cid))
+                     third_party_wrapper_window, cid), daemon=True)
     t.start()
     t.join()
     hwnd = get_handle_id(third_party_application_title)
+    logging.debug('[Nest Third Party Application] Client Hwnd = {hwnd}'.format(hwnd=hwnd))
     if hwnd != 0:
+        third_party_client_launch_status[app_tag] = 1
         if cid != '':
             BrowserView.cid_hwnd_map[cid] = hwnd
             BrowserView.hwnd_subprocess_map[hwnd] = BrowserView.cid_3rd_client_process_map[cid]
@@ -976,8 +1100,6 @@ def nest_third_party_application(target_uid='master',
         temp_window = 0
         while temp_window == 0:
             temp_window = QWindow.fromWinId(hwnd)
-        logging.debug('temp_window = {temp_window}'.format(temp_window=temp_window))
-        logging.debug('hwnd = {hwnd}'.format(hwnd=hwnd))
         temp_window.setFlags(
             Qt.CustomizeWindowHint | Qt.WindowTitleHint | Qt.FramelessWindowHint | Qt.WA_TranslucentBackground)
         third_party_window = ThirdPartyWindow(temp_window)
@@ -1034,19 +1156,47 @@ def generate_guid(prefix='child'):
 
 def open_new_window(url, title=default_window_title, payload=None, maximized=False, minimized=False, cid='',
                     width=default_window_width, height=default_window_height, enable_max=True):
+    logging.debug('[Open New Window] : url={url}, cid={cid}'.format(url=url, cid=cid))
     create_browser_view(uid=generate_guid(), url=url, title=title, payload=payload, maximized=maximized,
                         minimized=minimized, cid=cid, width=width, height=height, enable_max=enable_max,
                         inject_obj=app_inject_obj, application_version=app_version_info)
 
 
-def create_qt_view(default_show=True, window_type='cef', url='', cid='', payload={}):
-    uid = generate_guid()
-    qt_view = BrowserView(uid, window_type=window_type, url=url, cid=cid)
-    if window_type == 'cef':
-        set_client_handler(uid, payload=payload, cid=cid, browser=qt_view)
-        set_javascript_bindings(uid)
+def create_qt_view(default_show=True, window_type='cef', url='', cid='', payload={}, popup_validate=None):
+    """
+    基于Qt创建新窗口
+    :param default_show:  默认是否显示窗口
+    :param window_type:  窗口类型：cef => 浏览器窗口 , qt => qt窗口
+    :param url:  创建浏览器窗口时候，可以指定默认打开链接地址
+    :param cid:  创建的新窗口的customer window id
+    :param payload:  创建浏览器新窗口的时候需要向窗口注入的挂载数据
+    :param popup_validate:  当创建窗口为浏览器窗口的时候，设置此窗口关于弹框相关的校验逻辑
+    :return:
+    """
+    # print('[create_qt_window]')
+    # print('cid = ', cid)
+    # print('BrowserView.cid_map = ', BrowserView.cid_map)
+    # print('window_type = ', window_type)
+    qt_view = None
+    has_exist = False  # 是否已经存在cid相同的window
+    for (UID, CID) in BrowserView.cid_map.items():
+        if CID == cid:
+            has_exist = True
+            qt_view = BrowserView.instances[UID]
+            break
+
+    if not qt_view:
+        uid = generate_guid()
+        qt_view = BrowserView(uid, window_type=window_type, url=url, cid=cid, popup_validate=popup_validate)
+        qt_view.update_browser_info_one_by_one()
+
+    if window_type == 'cef' and not has_exist:
+        set_javascript_bindings(uid, payload=payload)
+        set_client_handler(uid, payload=payload, cid=cid, browser=qt_view, popup_validate=popup_validate)
+
     if default_show:
         qt_view.show()
+
     return qt_view
 
 
@@ -1055,10 +1205,12 @@ def create_browser_view(uid, title="", url=None, width=default_window_width, hei
                         min_size=(min_window_width, min_window_height),
                         background_color="#ffffff", web_view_ready=None, payload=None, maximized=False,
                         minimized=False, cid='', call_back=None, enable_max=True, inject_obj=app_inject_obj,
-                        quit_app_when_main_window_close=True, application_version=app_version_info):
+                        quit_app_when_main_window_close=True, application_version=app_version_info,
+                        popup_validate=None):
     browser = BrowserView(uid, title, url, width, height, resizable, full_screen, min_size,
                           background_color, web_view_ready, cid=cid, enable_max=enable_max,
-                          quit_app_when_main_window_close=quit_app_when_main_window_close)
+                          quit_app_when_main_window_close=quit_app_when_main_window_close,
+                          popup_validate=popup_validate)
     if maximized:
         browser.showMaximized()
 
@@ -1066,11 +1218,15 @@ def create_browser_view(uid, title="", url=None, width=default_window_width, hei
         browser.showMinimized()
 
     browser.show()
+
+    global splash
+    if splash is not None:
+        splash.finish(browser)
+
     if hasattr(call_back, '__call__'):
         call_back(browser)
-
-    set_client_handler(uid, payload, cid, browser)
-    set_javascript_bindings(uid, inject_obj=inject_obj, application_version=application_version)
+    set_javascript_bindings(uid, inject_obj=inject_obj, application_version=application_version, payload=payload)
+    set_client_handler(uid, payload, cid, browser, popup_validate)
 
 
 def launch_main_window(uid, title, url, width, height, resizable, full_screen, min_size,
@@ -1078,14 +1234,10 @@ def launch_main_window(uid, title, url, width, height, resizable, full_screen, m
                        user_agent=None, icon_path='', call_back=None, single_instance=True,
                        app_name="Qt Application Made By Burgeon Developer", inject_obj=app_inject_obj,
                        quit_app_when_main_window_close=True, application_version=app_version_info,
-                       enable_cef_cache=True, main_window_class_name='Qt5QWindowIcon'):
-    hwnd = find_main_window_hwnd(class_name=main_window_class_name, title=title)
-    if single_instance and hwnd != 0 and platform.system() == 'Windows':
-        import win32gui, win32con
-        win32gui.ShowWindow(hwnd, win32con.SW_RESTORE)
-        win32gui.SetForegroundWindow(hwnd)
-        exit_python()
-        return
+                       enable_cef_cache=True, main_window_class_name=qt_class_name, popup_validate=None,
+                       splash_screen=False, splash_text='',
+                       splash_style='font-size: 20px; font-family: Microsoft YaHei;'):
+    logging.debug('\r\n ---------- Application will launch ----------')
     global app_inject_obj
     global app_version_info
     global global_icon_path
@@ -1095,6 +1247,36 @@ def launch_main_window(uid, title, url, width, height, resizable, full_screen, m
     global app
     app = CefApplication(sys.argv)
     app.setApplicationName(app_name)
+
+    # 第一层拦截，应用单实例层面拦截
+    singleton = QSharedMemory(app.applicationName())
+    logging.debug('Application Pid = {pid} , Application has not launched before = {tag}'.format(pid=os.getpid(),
+                                                                                                 tag=singleton.create(
+                                                                                                     1)))
+    if platform.system() == 'Windows' and single_instance and not singleton.create(1):
+        # 第二层拦截，如果找到窗口句柄，则唤起
+        hwnd = find_main_window_hwnd(class_name=main_window_class_name, title=title)
+        if single_instance and hwnd != 0 and platform.system() == 'Windows':
+            import win32gui, win32con
+            logging.debug('[ pid = {pid} ] 唤起已经打开应用'.format(pid=os.getpid()))
+            win32gui.ShowWindow(hwnd, win32con.SW_RESTORE)
+            win32gui.SetForegroundWindow(hwnd)
+            logging.debug('[ pid = {pid} ]【退出本次启动】'.format(pid=os.getpid()))
+            exit_python()
+            return
+        exit_python()
+        return
+
+    # show splash 启动界面控制
+    if splash_screen:
+        global splash
+        splash = QSplashScreen()
+        splash.showMessage(splash_text, Qt.AlignHCenter | Qt.AlignVCenter | Qt.AlignCenter | Qt.AlignTop)
+        splash.setStyleSheet(splash_style)
+        splash.resize(300, 185.4)
+        splash.move(QApplication.desktop().availableGeometry().center() - splash.rect().center())
+        splash.show()
+
     cef_work_path = ''
     if platform.system() == 'Windows':
         cef_work_path = os.environ['ALLUSERSPROFILE']
@@ -1126,7 +1308,7 @@ def launch_main_window(uid, title, url, width, height, resizable, full_screen, m
                         background_color=background_color, web_view_ready=web_view_ready, maximized=maximized,
                         minimized=minimized, call_back=call_back, inject_obj=inject_obj,
                         quit_app_when_main_window_close=quit_app_when_main_window_close,
-                        application_version=application_version)
+                        application_version=application_version, popup_validate=popup_validate)
     logging.debug('Application Main Window Launched.')
     app.exec_()
     app.stop_timer()
@@ -1134,26 +1316,35 @@ def launch_main_window(uid, title, url, width, height, resizable, full_screen, m
     cef.Shutdown()
 
 
-def set_client_handler(uid, payload, cid, browser):
+def set_client_handler(uid, payload, cid, browser, popup_validate):
     BrowserView.instances[uid].view.SetClientHandler(LoadHandler(uid, payload, cid, browser))
+    BrowserView.instances[uid].view.SetClientHandler(LifeSpanHandler(popup_validate))
 
 
-def set_javascript_bindings(uid, inject_obj=app_inject_obj, application_version=app_version_info):
+def set_javascript_bindings(uid, inject_obj=app_inject_obj, application_version=app_version_info, payload={}):
     bindings = cef.JavascriptBindings(bindToFrames=False, bindToPopups=False)
     bindings.SetProperty('clientWorkDirectory', os.getcwd())
     bindings.SetProperty('currentClientVersion', application_version)
     bindings.SetProperty("cefPython3", cef.GetVersion())
     bindings.SetProperty('cefCachePath', app_cef_cache_path)
     bindings.SetProperty('windowId', uid)
+    bindings.SetProperty('qt5CefVersion', constant.qt5_cef_version)
     bindings.SetProperty('system', platform.system())
     bindings.SetProperty('systemLanguage', get_system_language())
+    bindings.SetProperty('__cef__', {
+        'cid': BrowserView.instances[uid].cid,
+        'cidLists': list(BrowserView.cid_map.values()),
+        'wid': uid,
+        'widLists': list(BrowserView.cid_map.keys()),
+        'payload': payload
+    })
     bindings.SetObject('windowInstance', BrowserView.instances[uid])
     for (k, v) in inject_obj.items():
         try:
             v.browser = BrowserView.instances[uid].view
         except Exception as e:
             logging.debug(e)
-        if v is None or isinstance(v, str):
+        if v is None or isinstance(v, str) or isinstance(v, dict):
             bindings.SetProperty(k, v)
         else:
             bindings.SetObject(k, v)
@@ -1161,6 +1352,9 @@ def set_javascript_bindings(uid, inject_obj=app_inject_obj, application_version=
 
 
 def append_payload(uid, payload, cid=''):
+    if not BrowserView.instances.__contains__(uid):
+        return
+
     BrowserView.instances[uid].view.ExecuteFunction('window.python_cef.updateCefConfig', 'wid', uid)
     if cid != '':
         BrowserView.instances[uid].view.ExecuteFunction('window.python_cef.updateCefConfig', 'cid', cid)
@@ -1198,6 +1392,11 @@ def set_cookies(url, cookies):
     cookie_manager.SetCookie(url, cookies)
 
 
+def delete_cookies(url, name):
+    cookie_manager = cef.CookieManager().GetGlobalManager()
+    cookie_manager.DeleteCookies(url, name)
+
+
 def quit_application():
     app.quit()
     t = Thread(target=exit_python)
